/*
 *  main.cc -- BabyTrans ( Babylon Translator front-end for GTK )
 *
 *  Copyright (C) 1999-2002  Frederic Jolliton -- <babytrans@tuxee.net>
 *  Applet changes Copyright (C) 2000 by Marcus Schiesser
 *  <marcus.schiesser@student.uni-tuebingen.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#else
# error "HAVE_CONFIG_H not set"
#endif

#include <iostream>
#include <string>
#include <vector>
#include <cerrno>

#include <sys/stat.h> // for stat()
#include <unistd.h> // for stat()
#include <getopt.h>

#include "fred_sstream"

// applet stuff
#ifdef GNOME_APPLET
# include <applet-widget.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "libbab.h"
#include "main.h"
#include "str_format.h"
#include "file_parser.h"
#include "gui_info.h"
#include "gui_main.h"
#include "misc.h"
#include "common.h"

//--[ namespace app ]----------------------------------------------------------

namespace {

/* The default filename for the home dictionary config */
std::string			defaultHomeDictionaryConfig = "~/.babytrans/dictionary" ;

/* The default filename for the home configuration */
std::string			defaultHomeConfig = "~/.babytrans/config" ;

std::string			prefsLanguage ;     /* The current language name */ 

Babylon				translator ;        /* The translator instance */

GtkToolbarStyle		prefsToolbarStyle ; /* The current toolbar style */

/* If true then automatically translate the first word in the current primary
 * X Window selection */
bool				prefsPeriodic = true ;

/* If true translate the word in the current primary X Window selection when
 * the program start. */
bool				immediate ;

/* Contain the first word to translate when program * start.  */
std::string			immediateWord ;

/* The default path to search for configuration file */
std::string			defaultPath ;

/* ID of the GTK+ timeout used to check if new word are selected for
 * translation */
guint				timeoutId = 0 ;

/* The last translated word. Used to check against the current selected word to
 * determine if new translation must be displayed */
std::string			lastWord ;

/* If true then translate new word when they are selected. */
bool				autoMode ;

/* FIXME: This don't really check the file existence.. */
bool
fileExist( const std::string& filename )
{
	struct stat			buf ;
	if ( stat( filename.c_str() , &buf ) == 0 )
		return true ;
	return false ;
}

/* Return environment variable as a std::string */
std::string
getenvString( const std::string& str )
{
	const char*			p = getenv( str.c_str() ) ;
	if ( p == NULL )
		return std::string() ;
	return std::string( p ) ;
}

/* expand ^(~)/.*$ or ^(~)$ to $HOME */
std::string
expandPathname( const std::string& path )
{
	std::string				home = getenvString( "HOME" ) ;
	if ( path.empty() || home.empty() )
		return path ;
	if ( path[ 0 ] != '~' )
		return path ;
	if ( path.size() >= 1 && path[ 1 ] != '/' )
		return path ;
	return home + path.substr( 1 ) ;
}

/* Call translateSelection() periodically */
gboolean
autoCbTimeout( gpointer data )
{
	guiMain::translateSelection() ;
	return TRUE ;
}

void
printHelp( const std::string& name )
{
	std::cout << PACKAGE " " VERSION "\n" ;
	// %1 is the application name
	std::cout << strFormat( _(
"Usage: %1 [OPTION]...\n"
"GTK+ front-end for Babylon Translator dictionaries\n\n"
"  -i, --immediate          Translate selection when program start.\n"
"  -l, --language=LANGUAGE  Specify the default language to use.\n"
"  -h, --help               Display this help, then exit.\n"
"  -p, --periodic           Periodically translate the selection.\n"
"  -V, --version            Display version number, then exit.\n"
"  -w, --word=WORD          Specify a word to translate when program start.\n"
"\n"
"Report bugs to babytrans@tuxee.net\n" ) ).with( name ) ;
}

option
longOption[] = {
#ifdef GNOME_APPLET
	{ "applet"    , 0 , 0 , 'a' } ,
#endif
	{ "immediate" , 0 , 0 , 'i' } ,
	{ "help"      , 0 , 0 , 'h' } ,
	{ "language"  , 1 , 0 , 'l' } ,
	{ "periodic"  , 0 , 0 , 'p' } ,
	{ "version"   , 0 , 0 , 'V' } ,
	{ "word"      , 1 , 0 , 'w' } ,
	{ 0           , 0 , 0 , 0 }
} ;

int
parseCommandLine( int argc , char** argv )
{
	int					optionIndex = 0 ;
	int					rc = 0 ;
	if ( argc == 0 ) { return -1 ; }
	std::string			name = argv[ 0 ] ;
	while ( rc == 0 ) {
		int					c = getopt_long( argc , argv , "aihl:pVw:" ,
						 					 longOption , &optionIndex ) ;
		if ( c == -1 ) { break ; }
		switch ( c ) {
			case 'a' :
				// FIXME
				break ;
			case 'i' :
				immediate = true ;
				break ;
			case 'h' :
				printHelp( argv[ 0 ] ) ;
				rc = 1 ;
				break ;
			case 'l' :
				prefsLanguage = std::string( optarg ) ;
				break ;
			case 'p' :
				prefsPeriodic = true ;
				break ;
			case 'V' :
				std::cout << PACKAGE " " VERSION "\n" ;
				rc = 1 ;
				break ;
			case 'w' :
				immediateWord = std::string( optarg ) ;
				break ;
			default :
				rc = -1 ;
				break ;
		}
	}
	return rc ;
}

int
dictionaryNameToIndex( const std::string& language )
{
	for ( app::LanguageList::size_type
		  i = 0 ;
		  i < app::languages.size() ;
		  ++ i ) {
		if ( app::languages[ i ].shortName == language )
			return int( i ) ;
	}
	return -1 ;
}

std::string
decodeAttribute( const std::string& attrib )
{
	std::string			humanAttrib ;
	assert( attrib.size() >= 8 ) ;
	if ( attrib[ 2 ] > '1' || attrib[ 3 ] != '0' ) {
		/* noun */
		humanAttrib = _( "n." ) ;
	} else if ( attrib[ 6 ] == '1' ) {
		/* adverb */
		humanAttrib = _( "adv." ) ;
	} else if ( attrib[ 6 ] != '0' || attrib[ 9 ] != '0' ) {
		/* verb */
		humanAttrib = _( "v." ) ;
	} else if ( attrib[ 7 ] != '0' ) {
		/* adverb */
		humanAttrib = _( "adv." ) ;
	} else if ( attrib[ 8 ] == '1' ) {
		/* adjective */
		humanAttrib = _( "a." ) ;
	} else
		humanAttrib = attrib ;

	return humanAttrib ;
}

bool
loadPreferences( const std::string& prefsFilename )
{
	std::ifstream		preferences( prefsFilename.c_str() ) ;
	if ( ! preferences ) {
		std::cerr << prefsFilename << ": " << strerror( errno ) << std::endl ;
		return false ;
	}
	std::vector< std::string >	vs ;
	size_t				line = 0 ;
	FileParserStatus	rc = fileParserOk ;
	while ( rc == fileParserOk ) {
		rc = readConfigLine( preferences , vs ) ;
		++ line ;
		if ( rc == fileParserOk && ! vs.empty() ) {
			/* All line must be of the form: keyword=argument */
			if ( vs.size() != 2 ) {
				rc = fileParserSyntaxError ;
			} else if ( vs[ 0 ] == "toolbar_style" ) {
				if ( vs[ 1 ] == "text" ) {
					prefsToolbarStyle = GTK_TOOLBAR_TEXT ;
				} else if ( vs[ 1 ] == "icons" ) {
					prefsToolbarStyle = GTK_TOOLBAR_ICONS ;
				} else if ( vs[ 1 ] == "both" ) {
					prefsToolbarStyle = GTK_TOOLBAR_BOTH ;
				} else {
					rc = fileParserSyntaxError ;
				}
			} else if ( vs[ 0 ] == "language" ) {
				if ( prefsLanguage.empty() ) {
					prefsLanguage = vs[ 1 ] ;
				}
			} else if ( vs[ 0 ] == "auto" ) {
				if ( vs[ 1 ] == "true" ) {
					prefsPeriodic = true ;
				} else if ( vs[ 1 ] == "false" ) {
					// app::prefsPeriodic = false ;
				} else {
					rc = fileParserSyntaxError ;
				}
			} else if ( vs[ 0 ] == "auto_raise" ) {
				if ( vs[ 1 ] == "true" ) {
					app::autoRaise = true ;
				} else if ( vs[ 1 ] == "false" ) {
					app::autoRaise = false ;
				} else {
					rc = fileParserSyntaxError ;
				}
			} else {
				rc = fileParserSyntaxError ;
			}
		}
	}
	if ( rc == fileParserSyntaxError ) {
		std::cerr << prefsFilename << ':'
				  << line << ':'
				  << _( "syntax error" )
				  << std::endl ;
	} else if ( rc == fileParserIoError ) {
		std::cerr << strFormat( _( "** error while reading %1" ) )
						.with( prefsFilename ) << std::endl ;
	}

	return ( rc == fileParserEof ) ;
}

/* Read the dictionary list file and fill app::languages.
 */
bool
loadDictionaryList( const std::string& dict_filename )
{
	/* load dictionary list */
	std::ifstream		dictionaryConfig( dict_filename.c_str() ) ;
	if ( ! dictionaryConfig ) {
		std::cerr << dict_filename << ": " << strerror( errno ) << std::endl ;
		return false ;
	}
	std::vector< std::string >	vs ;
	size_t				line = 0 ;
	FileParserStatus	rc = fileParserOk ;
	while ( rc == fileParserOk ) {
		rc = readConfigLine( dictionaryConfig , vs ) ;
		++ line ;
		if ( rc == fileParserOk && ! vs.empty() ) {
			if ( vs[ 0 ] == "path" ) {
				/* 'path' = pathname */
				if ( vs.size() == 2 ) {
					translator.setPath( expandPathname( vs[ 1 ] ) ) ;
				} else {
					rc = fileParserSyntaxError ;
				}
			} else if ( vs.size() == 3 ) {
				/* language_shorcut = language_name , language_filename */
				app::languages.push_back( app::LanguageItem( vs[ 0 ] ,
															 vs[ 1 ] ,
															 vs[ 2 ] ) ) ;
			} else {
				rc = fileParserSyntaxError ;
			}
		}
	}
	if ( rc == fileParserSyntaxError ) {
		std::cerr << dict_filename << ':'
				  << line << ':'
				  << _( "syntax error" )
				  << std::endl ;
	} else if ( rc == fileParserSyntaxError ) {
		std::cerr << strFormat( _( "** error while reading %1" ) )
			.with( dict_filename ) << std::endl ;
	}

	return ( rc == fileParserEof ) ;
}

/* Read configuration files.
 */
bool
loadSetup()
{
	/*
	 * Read the dictionaries list
	 */
	std::string			homeDictConfig =
		expandPathname( defaultHomeDictionaryConfig ) ;
	std::string			globalDictConfig = defaultPath + "/dictionary" ;

	if ( fileExist( homeDictConfig ) ) {
		/* Try to read the personnal dictionaries list first */
		if ( ! loadDictionaryList( homeDictConfig ) )
			return false ;
	} else {
		/* Read the system-wide dictionaries list */
		if ( ! loadDictionaryList( globalDictConfig ) )
			return false ;
	}
	if ( app::languages.empty() ) {
		std::cerr << _( "File describing available dictionaries was not found.\n"
			"Please create or copy dictionaries list in one of these files:\n" ) ;
		std::cerr << "  " << homeDictConfig << std::endl ;
		std::cerr << "  " << defaultPath + "/dictionary" << std::endl ;
		std::cerr << _( "Check the documentation for the format to use, or where\n"
			"to find a default configuration file\n" ) ;
		return false ;
	}

	/*
	 * Read the configuration file
	 */
	std::string				homeConfig = expandPathname( defaultHomeConfig ) ;
	std::string				globalConfig = defaultPath + "/config" ;
	bool				ok = false ;

	if ( fileExist( homeConfig ) )
		/* Try to read the personnal configuration file */
		ok = loadPreferences( homeConfig ) ;
	else if ( fileExist( globalConfig ) )
		/* Read the system-wide configuration file */
		ok = loadPreferences( globalConfig ) ;
	else
		ok = true ;

	return ok ;
}

/*
 * Update configuration file
 */
void
writeSetup()
{
	/* Create ~/.babytrans directory */
	mkdir( expandPathname( "~/.babytrans" ).c_str() , 0777 ) ;

	std::ofstream		config( expandPathname( "~/.babytrans/config" )
								.c_str() ) ;
	if ( config.is_open() ) {
		// %1 is the application name
		config << strFormat( _( "# Created by %1\n#\n"
			"# toolbar_style = [text|icons|both]\n"
			"# language = language_abbreviation (de,fr,es,...) that must\n"
			"# correspond with entry in 'dictionary' file\n"
			"# auto = [false|true]\n"
			"# autoRaise = [false|true]\n\n" ) )
			.with( PACKAGE " " VERSION ) ;
		config << "toolbar_style = " ;
		switch( guiMain::getToolbarStyle() ) {
		case GTK_TOOLBAR_TEXT:
			config << "text" ;
			break ;
		case GTK_TOOLBAR_ICONS:
			config << "icons" ;
			break ;
		default:
			config << "both" ;
			break ;
		}
		config << "\nlanguage = "
				<< app::languages[ app::currentDictionaryIndex ].shortName
				<< '\n' ;
		config << "auto = "
				<< ( ( timeoutId == 0 ) ? "false" : "true" )
				<< '\n' ;
		config << "auto_raise = "
				<< ( app::autoRaise ? "true" : "false" )
				<< '\n' ;
		config.close() ;
	}
}

} // anonymous namespace

app::LanguageList	app::languages ; /* The list of dictionary available */

/* Index in 'languages' for the current selected dictionary used
 * for translation */
int					app::currentDictionaryIndex ;

/* If true then raise (and uniconify) the window when
 * new translation occur
 */
bool				app::autoRaise ;

/* true if babytrans run as an applet */
bool				app::applet ;

/*
 * Set the current dictionary used for translation,
 * then update window.
 */
bool
app::setDictionary( const std::string& language )
{
	static const char*  strError = 0 ;
	if ( strError == 0 ) { strError = _( "E R R O R" ) ; }

	int					i = dictionaryNameToIndex( language ) ;		
	bool				rv = false ;
	if ( i >= 0 ) {
		rv = translator.open( app::languages[ i ].filename ) ;
		if ( rv ) { /* Ok */
			lastWord.erase() ; /* Invalidate lastWord to force translation */
			currentDictionaryIndex = i ;
			guiMain::update() ;
			app::translateWord( guiMain::getTextEntry() ) ;
		} else { /* Not found or other error */
			guiInfo::display( strError ,
							  translator.getError() , _( "OK" ) ) ;
			// Back to the previous dictionary
			rv = translator.open(
				app::languages[ currentDictionaryIndex ].filename ) ;
		}
	} else {
		/* Unknown language */
		guiInfo::display( strError ,
						  strFormat( _( "Unknown dictionary '%1'" ) )
							.with( language ) ,
						  _( "OK" ) ) ;
	}
	return rv ;
}

int
app::getDictionary()
{
	return app::currentDictionaryIndex ;
}

bool
app::getAuto()
{
	return autoMode ;
}

void
app::setAuto( bool flag )
{
	autoMode = flag ;
	if ( flag ) {
		if ( timeoutId == 0 )
			timeoutId = gtk_timeout_add( 2 * 1000 , autoCbTimeout , 0 ) ;
	} else {
		if ( timeoutId != 0 ) {
			gtk_timeout_remove( timeoutId ) ;
			timeoutId = 0 ;
		}
	}
}

/*
 * Update the translation
 */
void
app::translateWord( const std::string& word )
{
	// Do nothing if word is unchanged
	if ( word == lastWord ) return ;

	lastWord = word ;

	guiMain::setTextEntry( word ) ;

	Babylon::ContainerType
						translations ;

	guiMain::clearText() ;

	/* Translate the word
	 *
	 * (All translations will be stored in 'translations' container)
	 */
	bool				ok = translator.translate( word , translations ) ;

	if ( ! ok ) {
		/* A error occured while translating word */
		guiMain::insertText( translator.getError() , guiMain::colorError ) ;
	} else if ( translations.empty() ) {
		/* No translation was found */
		guiMain::insertText( strFormat( "%1:\n" ).with( word ) ,
							 guiMain::colorHighlight ) ;
		guiMain::insertText( _( "No translation found." ) , guiMain::colorError ) ;
	} else {
		/* Freeze the text widget to avoid lot of update from GTK+ while
		 * updating its contents.
		 */
		guiMain::freezeText() ;

		bool				firstTime = true ;
		for ( Babylon::ContainerType::iterator
			it = translations.begin() ;
			it != translations.end() ;
			++ it ) {
			if ( ! firstTime ) {
				guiMain::insertText( "\n\n" ) ;
			} else {
				firstTime = false ;
			}
			std::string			def_text = it->definition ;
			std::string::size_type
								i ;
#ifndef GNOME_APPLET
			/* If not running as a Gnome applet then
			 * go to new line instead of printing ';'
			 */
			for ( i = 0 ; i < def_text.size() ; ++ i ) {
				if ( def_text[ i ] == ';' ) def_text[ i ] = '\n' ;
			}
#else
			/* If running as Gnome applet then
			 * keep ';' characters and append a space after each
			 * of them if necessary.
			 */
			for ( i = 0 ; i < def_text.size() ; ++ i ) {
				if ( def_text[ i ] == ';' )
					if ( def_text[ i + 1 ] != ' ' )
						def_text.insert( def_text.begin() + i + 1 , ' ' ) ;
			}
#endif
			/* Append a space after each ',' characters if necessary.
			 */
			for ( i = 0 ; i < def_text.size() ; ++ i ) {
				if ( def_text[ i ] == ',' )
					if ( def_text[ i + 1 ] != ' ' )
						def_text.insert( def_text.begin() + i + 1 , ' ' ) ;
			}
	
			std::string			humanAttrib = decodeAttribute( it->attrib ) ;

			guiMain::insertText( strFormat( "%1 (%2):\n" )
								 .with( it->word )
								 .with( humanAttrib ) ,
								 guiMain::colorHighlight ) ;
			guiMain::insertText( def_text , guiMain::colorNormal ) ;
		}
		guiMain::thawText() ;
	
		// we dont want to change the focus if we are running as an applet
#ifndef GNOME_APPLET
		guiMain::grabFocus() ;
#endif
	}
	if ( app::autoRaise )
		guiMain::raiseAndFocus() ;
}

//------------------------------------------------------------------------------

int
main( int argc , char* argv[] )
{
#ifdef ENABLE_NLS
	setlocale( LC_ALL , "" ) ;
	bindtextdomain( PACKAGE , LOCALEDIR ) ;
	textdomain( PACKAGE ) ;
#endif

	/* default setting */

	immediate = false ;
	defaultPath = DATADIR "/babytrans/" ;
	prefsToolbarStyle = GTK_TOOLBAR_BOTH ;

#ifdef GNOME_APPLET
	app::applet = true ;
	applet_widget_init( PACKAGE , VERSION , argc , argv , NULL , 0 , NULL ) ;
#else
	app::applet = false ;
	gtk_init( &argc , &argv ) ;
#endif

	int rc ;
	rc = parseCommandLine( argc , argv ) ;
	if ( rc != 0 ) {
		return ( rc > 0 ) ? EXIT_SUCCESS : EXIT_FAILURE ;
	}

	translator.setPath( defaultPath ) ;

	if ( ! loadSetup() ) {
		guiInfo::display( _( "E R R O R" ) ,
			_( "Error while reading configuration file" ) ,
			_( "OK" ) ) ;
		return EXIT_FAILURE ;
	}

	if ( prefsLanguage.empty() ) {
		guiInfo::display( _( "E R R O R" ) ,
			_( "Go to preference box to select a language" ) ,
			_( "OK" ) ) ;
	} else {
		bool				rc = app::setDictionary( prefsLanguage ) ;
		if ( ! rc ) {
			return EXIT_FAILURE ;
		}
	}

	guiMain::setToolbarStyle( prefsToolbarStyle ) ;
	guiMain::display() ;
	
	app::setAuto( prefsPeriodic ) ;

	if ( ! immediateWord.empty() ) {
		app::translateWord( immediateWord ) ;
	} else if ( prefsPeriodic || immediate ) {
		guiMain::translateSelection() ;
	}

	guiMain::update() ;

#ifdef GNOME_APPLET
	applet_widget_gtk_main();
#else
	gtk_main() ;
#endif

	writeSetup() ;
}

// vim: set ts=4 sw=4:
